public class Spring implements ISaveGame
{
    Vector m_X = null, m_V = null, m_A = null, m_ForceFactor = null, m_Attenuation = null;
    float m_fRetainedTime;
    
    // this defines the simulation time interval
    // (remember that it affects the attenuation)
    final static float CONSTANT_TIME_GRAIN = 0.01f;

    protected float m_fInducedStateThreshold = 0.01f;
    protected boolean m_bIsInduced = false, 
        m_bInducedStateDirty = true;
    
    Spring()
    {
        // let's zero state members
        m_X = new Vector(0,0,0);
        m_V = new Vector(0,0,0);
        m_A = new Vector(0,0,0);
        m_fRetainedTime = 0.0f;
        // and set some default values for spring parameters
        m_ForceFactor = new Vector(30.0f, 30.0f, 30.0f);
        m_Attenuation = new Vector(0.9f, 0.9f, 0.9f);    
    }
    
    void ApplySpringForce()
    {
        // let's apply the spring force which is
        // of opposite direction to the current position
        m_A.fX = ( - m_X.fX * m_ForceFactor.fX );
        m_A.fY = ( - m_X.fY * m_ForceFactor.fY );
        m_A.fZ = ( - m_X.fZ * m_ForceFactor.fZ );
    }
    
    void InternalCalculate(float fTime)
    {
        ApplySpringForce();

        // calculate _V_elocity from _A_cceletation
        m_V.fX += ( m_A.fX * fTime );
        m_V.fY += ( m_A.fY * fTime );
        m_V.fZ += ( m_A.fZ * fTime );

        // calculate position (X) from _V_elocity
        m_X.fX += ( m_V.fX * fTime );
        m_X.fY += ( m_V.fY * fTime );
        m_X.fZ += ( m_V.fZ * fTime );

        // let's apply the attenuation now:
        m_V.fX *= m_Attenuation.fX;
        m_V.fY *= m_Attenuation.fY;
        m_V.fZ *= m_Attenuation.fZ;
    }
    
    void Calculate(float fTime)
    {
        if (fTime > 0)
        {
            // we need to simulate the spring over the time fTime

            // so we add it to the time that was retained from previous simulation
            m_fRetainedTime += fTime;

            // now let's repeat the simulation lasting the constant interval
            // appriopriate number of times
            while (m_fRetainedTime>=CONSTANT_TIME_GRAIN)
            {
                InternalCalculate(CONSTANT_TIME_GRAIN);
                m_fRetainedTime-=CONSTANT_TIME_GRAIN;
            }
        }
    }    
    
    Vector GetPosition() 
    { 
        return m_X; 
    }
    
    void SetPosition(Vector Pos) 
    { 
        m_X.Set(Pos);
        m_bInducedStateDirty = true;
    }

    Vector GetVelocity() 
    { 
        return m_V; 
    }
    
    void SetVelocity(Vector Vel) 
    { 
        m_V.Set(Vel);
        m_bInducedStateDirty = true;
    }

    void SetVelocity(float fX, float fY, float fZ) 
    { 
        m_V.fX = fX;
        m_V.fY = fY;
        m_V.fZ = fZ;
        m_bInducedStateDirty = true;
    }

    void AddVelocity(Vector Vel) 
    {
        m_V.Add(Vel);
        m_bInducedStateDirty = true;
    }
    
    void AddVelocity(float fX, float fY, float fZ) 
    {
        m_V.fX += fX;
        m_V.fY += fY;
        m_V.fZ += fZ;
        m_bInducedStateDirty = true;
    }

    void SetAttenuation(Vector Attenuation) 
    { 
        m_Attenuation.Set(Attenuation); 
    }
    
    Vector GetAttenuation() 
    { 
        return m_Attenuation; 
    }

    void SetForceFactor(Vector ForceFactor) 
    { 
        m_ForceFactor.Set(ForceFactor); 
        m_bInducedStateDirty = true;
    }
    
    Vector GetForceFactor() 
    { 
        return m_ForceFactor; 
    }
    
    void SetInducedStateThreshold(float fThreshold)
    {
        m_fInducedStateThreshold = fThreshold;
        m_bInducedStateDirty = true;
    }
    
    float GetInducedStateThreshold()
    {
        return m_fInducedStateThreshold;
    }
    
    boolean IsInduced()
    {
        if (m_bIsInduced || m_bInducedStateDirty)
        {
            m_bIsInduced = Math.abs(m_A.fX) >= m_fInducedStateThreshold ||
                           Math.abs(m_A.fY) >= m_fInducedStateThreshold ||
                           Math.abs(m_A.fZ) >= m_fInducedStateThreshold ||
                           Math.abs(m_V.fX) >= m_fInducedStateThreshold ||
                           Math.abs(m_V.fY) >= m_fInducedStateThreshold ||
                           Math.abs(m_V.fZ) >= m_fInducedStateThreshold ||
                           Math.abs(m_X.fX) >= m_fInducedStateThreshold ||
                           Math.abs(m_X.fY) >= m_fInducedStateThreshold ||
                           Math.abs(m_X.fZ) >= m_fInducedStateThreshold;
            m_bInducedStateDirty = false;
        }
        return m_bIsInduced;
    }
    
    
    //*********************************************************
    //					SaveGame
    //*********************************************************
    
    /** Metoda wczytujaca zapisane dane w Save gracza
     */	
    public void SGLoadChunk(FileChunk cParentFC)
    {	
        FileChunk.Log("[Spring.SGLoadChunk]\n");
        
        FileChunk cFC = GameObject.LoadChunk(cParentFC);   
        
        if (cFC.GetID() == ESGChunksChrome._CHK_SPRING)
        {				
            m_X = cFC.LoadVector();
            m_V = cFC.LoadVector();
            m_A = cFC.LoadVector();
            m_ForceFactor = cFC.LoadVector();
            m_Attenuation = cFC.LoadVector();
            m_fRetainedTime = cFC.LoadFloat();
            m_fInducedStateThreshold = cFC.LoadFloat();
            m_bInducedStateDirty = true; // to spowoduje odswiezenie m_bInduced przy zapytaniu IsInduced. W Load, bo nie musi byc pozniej, a Update moze sie nie zawolac...
        }
        
        cFC.delete();
    }
    
    /** Metoda zapisujaca dane do Save gracza
     */	
    public void SGSaveChunk(FileChunk cFCParent)
    {
        FileChunk.Log("[Spring.SGSaveChunk]\n");

        FileChunk cFC = GameObject.NewChunk(ESGChunksChrome._CHK_SPRING, cFCParent);
        
        cFC.SaveVector(m_X);
        cFC.SaveVector(m_V);
        cFC.SaveVector(m_A);
        cFC.SaveVector(m_ForceFactor);
        cFC.SaveVector(m_Attenuation);	
        cFC.SaveFloat(m_fRetainedTime);		
        cFC.SaveFloat(m_fInducedStateThreshold);

        cFC.delete();
    }	    				   
    
    public void SGUpdate() {}
}
